from metaworkspace.runtime.processing.jobs import job
from metaworkspace.runtime.workspace import mws
from metacity.core.mapper import build_overlay
from metacity.core.layout import build_layout

class JobMapping(job.Job):
    TYPE = 'mappinglayer'

    def __init__(self):
        super().__init__()
    
    def setup(self, job_dir, project, source, target, overlay):
        super().setup(job_dir)
        self.project: str = project
        self.source: str = source
        self.target: str = target
        self.overlay: str = overlay

    def serialize(self):
        data = super().serialize()    
        data['project'] = self.project    
        data['source'] = self.source
        data['target'] = self.target
        data['overlay'] = self.overlay
        return data
    
    def deserialize(self, data):
        self.project = data['project']
        self.source = data['source']
        self.target = data['target']
        self.overlay = data['overlay']

    def run(self):
        self.update_status('loading resources')
        proj = mws.get_project(self.project)
        overlay = proj.create_overlay(self.overlay)
        source = proj.get_layer(self.source)
        target = proj.get_layer(self.target)
        self.update_status('mapping')
        update = lambda it: self.update_status(it)
        build_overlay(overlay, source, target, update)
        overlay.persist()
        self.update_status('building layout')
        build_layout(proj)
        self.update_status('finished')






from datetime import datetime, timedelta
from typing import Optional
from fastapi import APIRouter, Depends, FastAPI, HTTPException, status
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from jose import JWTError, jwt
from passlib.context import CryptContext
from pydantic import BaseModel
from metaworkspace.runtime.workspace import mws

class Token(BaseModel):
    access_token: str
    token_type: str


class TokenData(BaseModel):
    username: Optional[str] = None


class User(BaseModel):
    username: str
    email: Optional[str] = None
    full_name: Optional[str] = None


ALGORITHM = "HS256"
ACCESS_TOKEN_EXPIRE_MINUTES = 60

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")
app = FastAPI()

async def create_access_token(data: dict, expires_delta: Optional[timedelta] = None):
    to_encode = data.copy()
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    to_encode.update({"exp": expire})
    encoded_jwt = jwt.encode(to_encode, mws.security.secret, algorithm=ALGORITHM)
    return encoded_jwt


async def get_current_user(token: str = Depends(oauth2_scheme)):
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    try:
        payload = jwt.decode(token, mws.security.secret, algorithms=[ALGORITHM])
        username: str = payload.get("sub")
        if username is None:
            raise credentials_exception
        token_data = TokenData(username=username)
    except JWTError:
        raise credentials_exception
    user = mws.security.get_user(token_data.username)
    if user is None:
        raise credentials_exception
    return user


async def get_current_active_user(current_user: User = Depends(get_current_user)):
    return current_user


router = APIRouter(prefix="/auth",
                   tags=["auth"])

@router.post("/token", response_model=Token)
async def login_for_access_token(form_data: OAuth2PasswordRequestForm = Depends()):
    user = mws.security.verify_user(form_data.username, form_data.password)
    if not user:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )
    access_token_expires = timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    access_token = await create_access_token(
        data={"sub": user["username"]}, expires_delta=access_token_expires
    )
    return {"access_token": access_token, "token_type": "bearer"}


@router.get("/user", response_model=User)
async def read_users_me(current_user: User = Depends(get_current_active_user)):
    return current_user



from metaworkspace.runtime.processing.jobs import job
from metaworkspace.runtime.workspace import mws
from metacity.core.styles.apply import apply_style

class JobApplyStyle(job.Job):
    TYPE = 'applystyle'

    def __init__(self):
        super().__init__()

    def setup(self, job_dir, project, style_name):
        super().setup(job_dir)
        self.project: str = project
        self.style_name: str = style_name

    def serialize(self):
        data = super().serialize()    
        data['project'] = self.project    
        data['style_name'] = self.style_name    
        return data

    def deserialize(self, data):
        self.project = data['project']
        self.style_name = data['style_name']

    def run(self):
        self.update_status("getting style")
        proj = mws.get_project(self.project)
        self.update_status("parsing style")
       
        try:
            apply_style(proj, self.style_name)
        except Exception as e:
            self.update_status(e)
            self.update_status("parsing failed") 
            return

        self.update_status("finished")



from metaworkspace.runtime.processing.jobs import job
from metaworkspace.runtime.workspace import mws
from metacity.core.layout import build_layout


class JobBuildLayout(job.Job):
    TYPE = 'buildlayout'

    def __init__(self):
        super().__init__()
    
    def setup(self, job_dir, project):
        super().setup(job_dir)
        self.project: str = project

    def serialize(self):
        data = super().serialize()    
        data['project'] = self.project    
        return data
    
    def deserialize(self, data):
        self.project = data['project']

    def run(self):
        proj = mws.get_project(self.project)
        self.update_status("building layout")
        build_layout(proj)
        self.update_status("finished")






import os
from typing import List

from fastapi import (APIRouter, Depends, File, Form, HTTPException, Response,
                     UploadFile, status)
from fastapi.responses import JSONResponse
from metacity.core.styles.style import Style
from metacity.core.styles.apply import parse
from metaworkspace.runtime.api.auth import User, get_current_active_user
from metaworkspace.runtime.processing.jobs.applystyle import JobApplyStyle
from metaworkspace.runtime.processing.jobs.buildlayout import JobBuildLayout
from metaworkspace.runtime.processing.jobs.loaddataset import JobLoadDataset
from metaworkspace.runtime.processing.jobs.mapping import JobMapping
from metaworkspace.runtime.workspace import mws
from pydantic import BaseModel

router = APIRouter(prefix="/api",
                   tags=["api"])


class ProjectData(BaseModel):
    name: str


class RenameProjectData(BaseModel):
    old: str
    new: str


class LayerData(BaseModel):
    project: str
    name: str


class RenameLayerData(BaseModel):
    project: str
    old: str
    new: str


class MappingLayerData(BaseModel):
    project: str
    source: str
    target: str
    overlay: str

class LogData(BaseModel):
    name: str


class StyleData(BaseModel):
    project: str
    name: str
    styles: str


class ActionStyleData(BaseModel):
    project: str
    name: str


class RenameStyleData(BaseModel):
    project: str
    old: str
    new: str


class MetaData(BaseModel):
    project: str
    layer: str
    oid: int


#### PROJECT

@router.get("/projects", response_class=JSONResponse)
def list_projects():
    projects = []
    for project in mws.project_names:
        projects.append(project)
    return JSONResponse(projects)


@router.post("/project")
def add_project(project: ProjectData, current_user: User = Depends(get_current_active_user)):
    for p in mws.project_names:
        if p == project.name:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail=f"Project {project.name} already exists"
            )

    mws.create_project(project.name)
    return Response(status_code=status.HTTP_201_CREATED) 


@router.post("/project/rename")
def rename_project(project: RenameProjectData, current_user: User = Depends(get_current_active_user)):
    if mws.rename_project(project.old, project.new):
        return Response(status_code=status.HTTP_200_OK) 
    else:
        raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail=f"Project {project.new} already exists"
            )


@router.delete("/project")
def delete_project(project: ProjectData, current_user: User = Depends(get_current_active_user)):
    mws.delete_project(project.name)
    return Response(status_code=status.HTTP_200_OK)


@router.post("/project/meta")
def get_metadata(meta: MetaData):
    prj = mws.get_project(meta.project)
    if prj is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Project {meta.project} not found"
        )

    layer = prj.get_layer(meta.layer, load_model=False)

    try:
        object = layer[meta.oid]
        return JSONResponse(object.meta)
    except:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Object {meta.oid} not found"
        )

#### LAYER

@router.post("/layers", response_class=JSONResponse)
def list_layers(project: ProjectData, current_user: User = Depends(get_current_active_user)):
    prj = mws.get_project(project.name)
    if prj is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Project not found"
        )

    print(prj)
    data = []
    #TODO HERE
    for l in prj.clayers(load_set=False):
        data.append({ 'name': l.name, 'size': l.size, 'type' : l.type, 'disabled': l.disabled })
    return JSONResponse(data) 


@router.post("/layer")
def add_layer(project: str = Form(...), 
              layer: str = Form(...), 
              files: List[UploadFile] = File(...),
              current_user: User = Depends(get_current_active_user)):
    job = JobLoadDataset()
    if layer is None:
        layer = os.path.splitext(files[0].filename)[0]
    job_dir = mws.generate_job_dir()
    job.setup(job_dir, project, layer, files)
    job.submit()

    return Response(status_code=status.HTTP_201_CREATED)


@router.post("/layer/rename")
def rename_layer(layer: RenameLayerData, current_user: User = Depends(get_current_active_user)):
    prj = mws.get_project(layer.project)
    if prj is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Project not found"
        )

    if prj.rename_layer(layer.old, layer.new):
        return Response(status_code=status.HTTP_200_OK) 
    else:
        raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail=f"Layer {layer.new} already exists"
            )


@router.post("/layer/disable")
def disable_layer(layer: LayerData, current_user: User = Depends(get_current_active_user)):
    prj = mws.get_project(layer.project)
    if prj is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Project not found"
        )

    try:
        lyr = prj.get_layer(layer.name, load_set=False)
    except:
        lyr = prj.get_overlay(layer.name)
        
    lyr.disabled = True
    lyr.export()
    return Response(status_code=status.HTTP_200_OK)


@router.post("/layer/enable")
def enable_layer(layer: LayerData, current_user: User = Depends(get_current_active_user)):
    prj = mws.get_project(layer.project)
    if prj is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Project not found"
        )

    try:
        lyr = prj.get_layer(layer.name, load_set=False)
    except:
        lyr = prj.get_overlay(layer.name)
        
    lyr.disabled = False
    lyr.export()
    return Response(status_code=status.HTTP_200_OK)


@router.delete("/layer")
def delete_layer(layer: LayerData, current_user: User = Depends(get_current_active_user)):
    prj = mws.get_project(layer.project)
    if prj is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Project not found"
        )

    prj.delete_layer(layer.name)
    return Response(status_code=status.HTTP_200_OK)


##### LOGS

@router.get("/jobs", response_class=JSONResponse)
def list_jobs(current_user: User = Depends(get_current_active_user)):
    data = mws.get_current_job_configs()
    return JSONResponse(data) 


@router.get("/logs", response_class=JSONResponse)
def list_logs(current_user: User = Depends(get_current_active_user)):
    data = mws.get_logs()
    return JSONResponse(data) 


@router.post("/log", response_class=JSONResponse)
def get_log(log: LogData, current_user: User = Depends(get_current_active_user)):
    data = mws.get_log(log.name)
    if data == None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Log not found"
        )

    return JSONResponse(data) 

##### ACTIONS

@router.post("/project/build")
def build_project(project: ProjectData, current_user: User = Depends(get_current_active_user)):

    job = JobBuildLayout()
    job_dir = mws.generate_job_dir()
    job.setup(job_dir, project.name)
    job.submit()
    
    return Response(status_code=status.HTTP_200_OK)

@router.post("/layer/map")
def map_layer(layer: MappingLayerData, current_user: User = Depends(get_current_active_user)):

    job = JobMapping()
    job_dir = mws.generate_job_dir()
    job.setup(job_dir, layer.project, layer.source, layer.target, layer.overlay)
    job.submit()
    
    return Response(status_code=status.HTTP_200_OK)


@router.post("/style/apply")
def apply_style(style: ActionStyleData, current_user: User = Depends(get_current_active_user)):
    job = JobApplyStyle()
    job_dir = mws.generate_job_dir()
    job.setup(job_dir, style.project, style.name)
    job.submit()
    return Response(status_code=status.HTTP_200_OK)


@router.post("/style/parse")
def parse_style(style: StyleData, current_user: User = Depends(get_current_active_user)):
    try:
        parse(style.styles)
    except Exception as e:
            return JSONResponse({
                'status': 'error',
                'error': str(e)
            }) 
    return Response(status_code=status.HTTP_200_OK)




#### STYLES
@router.post("/styles", response_class=JSONResponse)
def get_styles(project: ProjectData, current_user: User = Depends(get_current_active_user)):
    prj = mws.get_project(project.name)
    if prj is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Project not found"
        )

    styles = Style.list(prj)
    return JSONResponse(styles) 


@router.post("/style/create")
def create_style(style: ActionStyleData, current_user: User = Depends(get_current_active_user)):
    prj = mws.get_project(style.project)
    if prj is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Project not found"
        )

    if Style.create(prj, style.name):
        return Response(status_code=status.HTTP_200_OK)
    else:
        raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail=f"Style {style.name} already exists"
            )


@router.post("/style")
def get_style(style: ActionStyleData, current_user: User = Depends(get_current_active_user)):
    proj = mws.get_project(style.project)
    if proj is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Project not found"
        )

    sstyle = Style(proj, style.name)
    try:
        style_data = sstyle.style_file
    except:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Style not found"
        )   
    return JSONResponse({'style': style_data, 'name': style.name }) 


@router.post("/style/update", response_class=JSONResponse)
def update_style(style: StyleData, current_user: User = Depends(get_current_active_user)):
    proj = mws.get_project(style.project)
    if proj is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Project not found"
        )

    sstyle = Style(proj, style.name)
    sstyle.update(style.styles)
    return Response(status_code=status.HTTP_200_OK)


@router.delete("/style")
def delete_style(style: ActionStyleData, current_user: User = Depends(get_current_active_user)):
    proj = mws.get_project(style.project)
    if proj is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Project not found"
        )

    sstyle = Style(proj, style.name)
    sstyle.delete()
    return Response(status_code=status.HTTP_200_OK)


@router.post("/style/rename")
def rename_style(style: RenameStyleData, current_user: User = Depends(get_current_active_user)):
    proj = mws.get_project(style.project)
    if proj is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Project not found"
        )

    sstyle = Style(proj, style.old)
    if sstyle.rename(style.new):
        return Response(status_code=status.HTTP_200_OK)
    raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Style {style.old} already exists"
        )



import os
import shutil
import uuid
import json
from datetime import datetime


def write_json(filename, data):
    if os.path.exists(filename):
        os.remove(filename)

    with open(filename, 'w') as file:
        json.dump(data, file, indent=4)


def read_json(filename):
    with open(filename, 'r') as file:
        return json.load(file)


def file_exists(file):
    return os.path.exists(file)


def remove_dirtree(dir):
    if os.path.exists(dir):
        shutil.rmtree(dir)


def create_dir_if_not_exists(dir):
    if not os.path.exists(dir):
        os.makedirs(dir)


def recreate_dir(dir):
    if os.path.exists(dir):
        shutil.rmtree(dir)
    create_dir_if_not_exists(dir)


def projects_dir(workspace_path):
    return os.path.join(workspace_path, "projects")


def jobs_dir(workspace_path):
    return os.path.join(workspace_path, "jobs")


def logs_dir(workspace_path):
    return os.path.join(workspace_path, "logs")


def user_dir(workspace_path):
    return os.path.join(workspace_path, "users")


def user_file(user_dir, username):
    return os.path.join(user_dir, f"{username}.json")


def security_config(workspace_path):
    return os.path.join(workspace_path, f"security.json")


def server_log(workspace_path):
    return os.path.join(logs_dir(workspace_path), "server_log.txt")


def jobs_log(workspace_path):
    return os.path.join(logs_dir(workspace_path), "jobs_log.txt")


def log_contents(workspace_path, log_name: str):
    if log_name.endswith(".log") or log_name.endswith(".txt"):
        path = os.path.join(logs_dir(workspace_path), log_name)
    else:
        path = os.path.join(logs_dir(workspace_path), log_name + ".log")
    
    if file_exists(path):
        with open(path, 'r') as file:
            return file.read()
    return None


def process_file(workspace_path):
    return os.path.join(workspace_path, "pid")


def project_dir(workspace_path, project_name):
    return os.path.join(projects_dir(workspace_path), project_name)


def get_projects(workspace_path):
    projects = projects_dir(workspace_path)
    for project in os.listdir(projects):
        if project[0] != '.':
            yield os.path.join(projects, project)

def get_project_names(workspace_path):
    projects = projects_dir(workspace_path)
    for project in os.listdir(projects):
        if project[0] != '.':
            yield project

def current_jobs(workspace_path):
    jobs = jobs_dir(workspace_path)
    return [job for job in os.listdir(jobs)]

def current_logs(workspace_path):
    logs = logs_dir(workspace_path)
    return [logs for logs in os.listdir(logs)]

def generate_job_dir(workspace_path):
    jobs = jobs_dir(workspace_path)
    names = current_jobs(workspace_path)
    jid = str(uuid.uuid4())
    while jid in names:
        jid = str(uuid.uuid4())
    job_dir = os.path.join(jobs, jid)
    create_dir_if_not_exists(job_dir)
    return job_dir


def id_from_jobdir(jobdir):
    return os.path.basename(jobdir)


def current_job_configs(workspace_path):
    jobdir = jobs_dir(workspace_path)
    configs = []
    for job in os.listdir(jobdir):
        conf = os.path.join(jobdir, job, "job.json")
        if file_exists(conf):
            configs.append(read_json(conf))
    return configs


def recreate_workspace(workspace_path):
    recreate_dir(workspace_path)
    create_dir_if_not_exists(projects_dir(workspace_path))
    create_dir_if_not_exists(jobs_dir(workspace_path))
    create_dir_if_not_exists(logs_dir(workspace_path))
    create_dir_if_not_exists(user_dir(workspace_path))
    

def archive_logs(workspace_path):
    logs = logs_dir(workspace_path)
    if os.path.exists(logs):
        new_logs_name = datetime.now().strftime('%Y-%m-%d-%H-%M–%S')
        new_logs_path = os.path.join(workspace_path, f'logs-{new_logs_name}')
        os.rename(logs, new_logs_path)
    create_dir_if_not_exists(logs)


def rename(old, new):
    if file_exists(old) and not file_exists(new):
        os.rename(old, new)
        return True
    return False



import os
import shutil

from metacity.core.grid.grid import build_grid
from metacity.core.timeline import build_timeline
from metacity.core.layout import build_layout
from metacity.io.parse import parse
from metaworkspace.runtime.processing.jobs import job
from metaworkspace.runtime.workspace import mws


class JobLoadDataset(job.Job):
    TYPE = 'loaddataset'

    def __init__(self):
        super().__init__()
    
    def setup(self, job_dir, project, layer, files):
        super().setup(job_dir)
        lfiles = self.setup_resources(job_dir, files)
        self.project: str = project
        self.layer = layer
        self.files = lfiles

    def serialize(self):
        data = super().serialize()    
        data['project'] = self.project    
        data['layer'] = self.layer    
        data['files'] = self.files    
        return data
    
    def deserialize(self, data):
        self.project = data['project']
        self.layer = data['layer']
        self.files = data['files']

    def run(self):
        layer, proj = self.parse_files() 
        layer.persist()
        self.update_status("building grid")
        build_grid(layer)
        self.update_status("building timeline")
        build_timeline(layer)
        self.update_status("building layout")
        build_layout(proj)
        self.update_status("finished")

    def setup_resources(self, job_dir, files):
        lfiles = []
        for ufile in files:
            local_file_path = os.path.join(job_dir, ufile.filename)
            with open(local_file_path, "wb+") as local_file:
                shutil.copyfileobj(ufile.file, local_file)
            lfiles.append(local_file_path)
        return lfiles

    def parse_files(self):
        proj = mws.get_project(self.project)
        layer = proj.create_layer(self.layer)
        for i, file in enumerate(self.files):
            self.update_status(f"parsing files: {i}/{len(self.files)}")
            try:
                layer.add_source_file(file)
                objects = parse(file)
                for o in objects:
                    layer.add(o)
            except Exception as e:
                self.log(e)
        return layer, proj
        





from typing import Dict, Type
from metaworkspace.runtime.processing.jobs.job import Job
from metaworkspace.runtime.processing.jobs.applystyle import JobApplyStyle
from metaworkspace.runtime.processing.jobs.loaddataset import JobLoadDataset
from metaworkspace.runtime.processing.jobs.buildlayout import JobBuildLayout
from metaworkspace.runtime.processing.jobs.mapping import JobMapping
import json
import os
import uuid
from time import sleep


jobs: Dict[str, Type[Job]] = {
    JobLoadDataset.TYPE: JobLoadDataset,
    JobBuildLayout.TYPE: JobBuildLayout,
    JobApplyStyle.TYPE: JobApplyStyle,
    JobMapping.TYPE: JobMapping
}


def job_object_hook(data):
    if 'type' not in data:
        raise Exception(f"Job {data} missing job type")
    
    t = data['type']
    if t not in jobs:
        raise Exception(f"Uknown job type: {t}")

    job = jobs[t]()
    job.deserialize(data)
    return job
    
     
def load_job(job_dir):
    jobfile = os.path.join(job_dir, 'job.json')

    for _ in range(5):
        try:
            with open(jobfile, 'r') as f:
                job: Job = json.load(f, object_hook=job_object_hook)
            job.job_dir = job_dir
            job.setup_log(str(uuid.uuid1()))
            return job
        except Exception as e:
            print(f"Failed to load job: {e}")
            print(f"Retrying in 5 seconds")
            sleep(5)
            return None







import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="metaworkspace",
    packages=find_packages("."),
    version="0.2.3",
    description="Python-React application for visualization of urban data",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/MetacitySuite/Metacity-Workspace",
    author="Metacity",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires='>=3.8', # Minimum version requirement of the package
    install_requires=[
        "metacity>=0.2.2",
        "requests>=2.26.0",
        "twine>=3.4.2",
        "fastapi>=0.68.1",
        "uvicorn>=0.15.0",
        "aiofiles>=0.7.0",
        "jinja2>=3.0.1",
        "watchdog>=2.1.5",
        "bcrypt>=3.2.0",
        "python-jose>=3.3.0",
        "python-jwt>=2.0.1",
        "passlib>=1.7.4",
        "python-multipart>=0.0.5"
    ],
    zip_safe=False,
    include_package_data=True
)
